# -*- coding: utf-8 -*-

# Copyright (C) 2006 by Magnus Therning

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# History: First epilicious used Frank Timmermann's pydelicious module. It
# worked fine for a long time (up to and including release 0.9) but there were
# a few things that irritated me:
#  1. pydelicious includes more than epilicious needs, meaning that I had to
#     cut a bit to keep dependencies reasonable
#  2. when it broke it broke in strange ways
#  3. there are bugs in pydelicious, some so basic that I have serious doubts
#     about the quality of the entire module
# Given this and the fact that epilicious needs only a small part of the
# pydelicious functionality I decided to write a custom del.icio.us module for
# epilicious.

import urllib
import urllib2
import sys
import time
# ElementTree is a standard lib from Python 2.5
if sys.version_info[0:2] >= (2, 5):
    from xml.etree.ElementTree import parse as xml_parse
else:
    from elementtree.ElementTree import parse as xml_parse


def _handle_throttle(function):
    # Decorator to deal with the throttling (503) that might happen when
    # talking to del.icio.us.

    def decorate(*args, **kwargs):
        while 1:
            try:
                return function(*args, **kwargs)
            except urllib2.HTTPError, e:
                if e.code != 503:
                    raise e
                time.sleep(1)

    # It'd be a shame to lose the docstrings, wouldn't it?
    decorate.__doc__ = function.__doc__
    return decorate


class _DeliciousParser(object):
    # Class for parsing the responses from del.icio.us. All methods expect an
    # ElementTree.

    def __init__(self):
        pass

    def parse_result(self, xmltree):
        # Parses a result tree. Returns 1 if done, 0 otherwise.
        if xmltree.getroot().attrib.get('code', '') == 'done':
            return 1
        else:
            return 0

    def parse_posts(self, xmltree):
        # Parses a <posts> tree into a list of dictionaries.
        keys = [ 'href', 'description', 'hash', 'tag', 'time', 'extended', \
                'count', ]
        l = []
        for n in xmltree.findall('post'):
            e = dict([[k, n.attrib.get(k, '')] for k in keys])
            l.append(e)
        return l


_DWS_REALM = 'del.icio.us API'
_DWS_BASE_URI = 'https://api.del.icio.us/'
_DWS_API_URI = 'https://api.del.icio.us/v1/'
_DWS_USER_AGENT = 'epilicious/0.9 (magnus@therning.org)'


class Delicious(object):
    def __init__(self, user_name, password):
        self.__user = user_name
        self.__passwd = password

    def __do_request(self, method, **kwargs):
        # Do the request to del.icio.us. Pass off the result to
        # __parse_result().
        for k in kwargs:
            if isinstance(kwargs[k], unicode):
                kwargs[k] = kwargs[k].encode('utf-8')
        params = urllib.urlencode(kwargs)
        authinfo = urllib2.HTTPBasicAuthHandler()
        authinfo.add_password(realm=_DWS_REALM, uri=_DWS_BASE_URI, user=self.__user, passwd=self.__passwd)
        opener = urllib2.build_opener(authinfo)
        request = urllib2.Request(_DWS_API_URI + method + '?' + params)
        request.add_header('User-Agent', _DWS_USER_AGENT)
        return self.__parse_result(xml_parse(opener.open(request)))

    def __parse_result(self, xmltree):
        # Pass on the actual parsing to a _DeliciousParser instance.
        fname = 'parse_' + xmltree.getroot().tag
        return getattr(_DeliciousParser(), fname)(xmltree)

    def all_bookmarks(self, tag=''):
        '''Get all bookmarks from del.icio.us.

        tag: (required) limits the response to bookmarks with that tag

        Returns a list of dictionaries:
            [{'count': ..., 'description': ..., 'extended': ..., 'hash': ...,
              'href': ..., 'tag': ..., 'time': ...},
            ]
        '''
        return self.__do_request('posts/all', tag=tag)
    all_bookmarks = _handle_throttle(all_bookmarks)

    def delete_bookmark(self, url):
        '''Delete bookmark from del.icio.us.

        url: (required) the URL

        Returns 1 on success, 0 on failure.
        '''
        return self.__do_request('posts/delete', url=url)
    delete_bookmark = _handle_throttle(delete_bookmark)

    def add_bookmark(self, url, description, notes='', tags='', dt='', replace='yes', shared='yes'):
        '''Add a bookmark to del.icio.us.

        url: (required) the URL
        description: (required) description of the URL
        notes: notes for the URL (default "")
        tags: string of tags for the URL (separated by spaces, default "")
        replace: whether an already existing bookmark should be replaced ("yes"
                 or "no", default "yes")
        shared: whether the URL should be shared or not ("yes" or "no", default
                 "no")

        Returns 1 on success, 0 on failure.
        '''
        return self.__do_request('posts/add', url=url, \
                description=description, extended=notes, tags=tags,
                dt=dt, replace=replace, shared=shared)
    add_bookmark = _handle_throttle(add_bookmark)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Listens to location changes using Python

import epiphany;

def location_cb(embed, address):
	print 'New location: %s' % address

def attach_tab(window, tab):
	embed = tab.get_embed()
	sig = embed.connect('ge-location', location_cb)
	embed._python_sample_location_sig = sig

def detach_tab(window, tab):
	embed = tab.get_embed()
	sig = embed._python_sample_location_sig
	del embed._python_sample_location_sig
	embed.disconnect(sig)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

print "Sample Python extension"
print "Copyright © 2005 Adam Hooper"
print

def attach_window(window):
	print "attach_window"

def detach_window(window):
	print "detach_window"

def attach_tab(window, tab):
	print "attach_tab"

def detach_tab(window, tab):
	print "detach_tab"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  Copyright © 2005 Christian Persch
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
#  $Id: favicon.py 1412 2006-12-19 16:20:56Z plaes $

import epiphany;
import gnomevfs;

def net_stop_cb(embed, tab):
	if tab.get_icon_address() == None:
		try:
			uri = gnomevfs.URI(tab.get_address())
			single = epiphany.ephy_shell_get_default().get_embed_single()
			backend = single.get_backend_name();
			if uri.scheme == "http" or (uri.scheme == "https" and backend != "gecko-1.7") :
				url = uri.scheme + "://" + uri.host_name
				if uri.host_port != 0:
					url += ':%d' % uri.host_port;
				url += "/favicon.ico"
				tab.set_icon_address(url);
		except Exception:
			pass

def attach_tab(window, tab):
	embed = tab.get_embed()
	handler_id = embed.connect("net_stop", net_stop_cb, tab)
	tab._favicon_details = [ handler_id ]

def detach_tab(window, tab):
	[ handler_id ] = tab._favicon_details
	del tab._favicon_details
	embed = tab.get_embed()
	embed.disconnect(handler_id)



# -*- coding: utf-8 -*-

# Copyright (C) 2006 by Magnus Therning

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import libepilicious
import gtk, gtk.glade
import os.path

class ProgressBar:

    def __init__(self):
        self.gui = gtk.glade.XML(os.path.join(libepilicious.GLADE_DIR, \
                'epilicious-progress.glade'), domain='epilicious')
        self.gui.signal_autoconnect(self)

        self.dlg = self.gui.get_widget('dlgProgress')

    def show(self):
        self.dlg.show()

    def hide(self):
        self.dlg.hide()

    def step(self):
        images = ['imgStep1', 'imgStep2', 'imgStep3', 'imgStep4', 'imgStep5', 'imgStep6',]
        for i in images:
            img = self.gui.get_widget(i)
            img.show()
            while gtk.events_pending():
                gtk.main_iteration()
            yield True
            img.set_from_icon_name('gtk-apply', gtk.ICON_SIZE_MENU)
        self.gui.get_widget('btnClose').set_sensitive(True)
        while gtk.events_pending():
            gtk.main_iteration()
        yield True

    def failed(self):
        self.dlg.set_title(self.dlg.get_title() + ' FAILED!')
        self.gui.get_widget('btnClose').set_sensitive(True)

    ### signal handlers
    def on_btnClose_clicked(self, widget):
        self.hide()
        del(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Interactive Python-GTK Console
#   Copyright ©, 1998 James Henstridge <james@daa.com.au>
#   Copyright ©, 2005 Adam Hooper <adamh@densi.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# This module implements an interactive python session in a GTK window.  To
# start the session, use the gtk_console command.  Its specification is:
#   gtk_console(namespace, title)
# where namespace is a dictionary representing the namespace of the session,
#       title     is the title on the window and
#
# As well as the starting attributes in namespace, the session will also
# have access to the list __history__, which is the command history.

import sys, string, traceback

import gobject
import gtk
import pango
import gtk.keysyms

stdout = sys.stdout

if not hasattr(sys, 'ps1'): sys.ps1 = '>>> '
if not hasattr(sys, 'ps2'): sys.ps2 = '... '

# some functions to help recognise breaks between commands
def remQuotStr(s):
	'''Returns s with any quoted strings removed (leaving quote marks)'''
	r = ''
	inq = 0
	qt = ''
	prev = '_'
	while len(s):
		s0, s = s[0], s[1:]
		if inq and (s0 != qt or prev == '\\'):
			prev = s0
			continue
		prev = s0
		if s0 in '\'"':
			if inq:
				inq = 0
			else:
				inq = 1
				qt = s0
		r = r + s0
	return r

def bracketsBalanced(s):
	'''Returns true iff the brackets in s are balanced'''
	b = filter(lambda x: x in '([{', s)
	e = filter(lambda x: x in ')]}', s)
	return len(e) >= len(b)

class gtkoutfile:
	'''A fake output file object.  It sends output to a TK test widget,
	and if asked for a file number, returns one set on instance creation'''
	def __init__(self, w, fn, tag):
		self.__fn = fn
		self.__w = w
		self.__tag = tag
	def close(self): pass
	flush = close
	def fileno(self):    return self.__fn
	def isatty(self):    return 0
	def read(self, a):   return ''
	def readline(self):  return ''
	def readlines(self): return []
	def write(self, s):
		#stdout.write(str(self.__w.get_point()) + '\n')
		self.__w.text_insert(self.__tag, s)
	def writelines(self, l):
		self.__w.text_insert(self.__tag, l)
	def seek(self, a):   raise IOError, (29, 'Illegal seek')
	def tell(self):      raise IOError, (29, 'Illegal seek')
	truncate = tell

class Console(gtk.VBox):
	def __init__(self, namespace={}, quit_cb=None):
		gtk.VBox.__init__(self, spacing=12)
		self.set_border_width(12)

		self.quit_cb = quit_cb

		self.inp = gtk.HBox()
		self.pack_start(self.inp)
		self.inp.show()

		self.sw = gtk.ScrolledWindow()
		self.sw.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC);
		self.sw.set_shadow_type(gtk.SHADOW_IN)
		self.inp.pack_start(self.sw, padding=1)
		self.sw.show()

		self.text = gtk.TextView()
		self.text.modify_font(pango.FontDescription('Monospace'))
		self.text.set_editable(False)
		self.text.set_wrap_mode (gtk.WRAP_WORD)
		self.sw.add(self.text)
		self.text.show()

		buffer = self.text.get_buffer()
		self.normal = buffer.create_tag("normal")
		self.error = buffer.create_tag("error")
		self.error.set_property("foreground", "red")
		self.command = buffer.create_tag('command')
		self.command.set_property("foreground", "blue")
 
		self.inputbox = gtk.HBox(spacing=2)
		self.pack_end(self.inputbox, expand=False)
		self.inputbox.show()

		self.prompt = gtk.Label(sys.ps1)
		self.prompt.set_padding(2, 0)
		self.prompt.modify_font(pango.FontDescription('Monospace 10'))
		self.inputbox.pack_start(self.prompt, expand=False)
		self.prompt.show()

		self.line = gtk.Entry()
		self.line.modify_font(pango.FontDescription('Monospace'))
		self.line.connect("key_press_event", self.key_function)
		self.inputbox.pack_start(self.line, padding=2)
		self.line.show()

		self.namespace = namespace

		self.cmd = ''
		self.cmd2 = ''

		# set up hooks for standard output.
		self.stdout = gtkoutfile(self, sys.stdout.fileno(),
					 self.normal)
		try :
			# this will mostly fail on win32 ...
			self.stderr = gtkoutfile(self, sys.stderr.fileno(),
						 self.error)
		except :
			# ... but gtkoutfile is not using the fileno anyway
			self.stderr = gtkoutfile(self, -1, self.error)
		# set up command history
		self.history = ['']
		self.histpos = 0
		self.namespace['__history__'] = self.history

	def scroll_to_end (self):
		iter = self.text.get_buffer().get_end_iter()
		self.text.scroll_to_iter(iter, 0.0)
		return False  # don't requeue this handler

	def text_insert(self, tag, s):
		buffer = self.text.get_buffer()
		iter = buffer.get_end_iter()
		buffer.insert_with_tags (iter, s, tag)
		gobject.idle_add(self.scroll_to_end)

	def init(self):
		self.text.realize()
		self.text.fg = self.text.style.fg[gtk.STATE_NORMAL]
		self.text.bg = self.text.style.white

		self.line.grab_focus()

	def quit(self, *args):
		#self.destroy()
		if self.quit_cb: self.quit_cb()

	def key_function(self, entry, event):
		if event.keyval == gtk.keysyms.Return:
			self.line.emit_stop_by_name("key_press_event")
			self.eval()
		elif event.keyval == gtk.keysyms.Tab:
		        self.line.emit_stop_by_name("key_press_event")
			self.line.insert_text('\t')
			gobject.idle_add(self.focus_text)
		elif event.keyval in (gtk.keysyms.KP_Up, gtk.keysyms.Up) \
		     or (event.keyval in (gtk.keysyms.P, gtk.keysyms.p) and
			 event.state & gtk.gdk.CONTROL_MASK):
			self.line.emit_stop_by_name("key_press_event")
			self.historyUp()
			gobject.idle_add(self.focus_text)
		elif event.keyval in (gtk.keysyms.KP_Down, gtk.keysyms.Down) \
		     or (event.keyval in (gtk.keysyms.N, gtk.keysyms.n) and
			 event.state & gtk.gdk.CONTROL_MASK):
			self.line.emit_stop_by_name("key_press_event")
			self.historyDown()
			gobject.idle_add(self.focus_text)
		elif event.keyval in (gtk.keysyms.L, gtk.keysyms.l) and \
		     event.state & gtk.gdk.CONTROL_MASK:
			self.text.get_buffer().set_text('')
		elif event.keyval in (gtk.keysyms.D, gtk.keysyms.d) and \
		     event.state & gtk.gdk.CONTROL_MASK:
			self.line.emit_stop_by_name("key_press_event")
			self.ctrld()

	def focus_text(self):
		self.line.grab_focus()
		return False  # don't requeue this handler

	def ctrld(self):
		self.quit()
		pass

	def historyUp(self):
		if self.histpos > 0:
			l = self.line.get_text()
			if len(l) > 0 and l[0] == '\n': l = l[1:]
			if len(l) > 0 and l[-1] == '\n': l = l[:-1]
			self.history[self.histpos] = l
			self.histpos = self.histpos - 1
			self.line.set_text(self.history[self.histpos])
			
	def historyDown(self):
		if self.histpos < len(self.history) - 1:
			l = self.line.get_text()
			if len(l) > 0 and l[0] == '\n': l = l[1:]
			if len(l) > 0 and l[-1] == '\n': l = l[:-1]
			self.history[self.histpos] = l
			self.histpos = self.histpos + 1
			self.line.set_text(self.history[self.histpos])

	def eval(self):
		l = self.line.get_text() + '\n'
		if len(l) > 1 and l[0] == '\n': l = l[1:]
		self.histpos = len(self.history) - 1
		if len(l) > 0 and l[-1] == '\n':
			self.history[self.histpos] = l[:-1]
		else:
			self.history[self.histpos] = l
		self.line.set_text('')
		self.text_insert(self.command, self.prompt.get() + l)
		if l == '\n':
			self.run(self.cmd)
			self.cmd = ''
			self.cmd2 = ''
			return
		self.histpos = self.histpos + 1
		self.history.append('')
		self.cmd = self.cmd + l
		self.cmd2 = self.cmd2 + remQuotStr(l)
		l = string.rstrip(l)
		if not bracketsBalanced(self.cmd2) or l[-1] == ':' or \
				l[-1] == '\\' or l[0] in ' \11':
			self.prompt.set_text(sys.ps2)
			self.prompt.queue_draw()
			return
		self.run(self.cmd)
		self.cmd = ''
		self.cmd2 = ''

	def run(self, cmd):
		sys.stdout, self.stdout = self.stdout, sys.stdout
		sys.stderr, self.stderr = self.stderr, sys.stderr
		try:
			try:
				r = eval(cmd, self.namespace, self.namespace)
				if r is not None:
					print `r`
			except SyntaxError:
				exec cmd in self.namespace
		except:
			if hasattr(sys, 'last_type') and \
					sys.last_type == SystemExit:
				self.quit()
			else:
				traceback.print_exc()
		self.prompt.set_text(sys.ps1)
		self.prompt.queue_draw()
		#adj = self.text.get_vadjustment()
		#adj.set_value(adj.upper - adj.page_size)
		sys.stdout, self.stdout = self.stdout, sys.stdout
		sys.stderr, self.stderr = self.stderr, sys.stderr

def gtk_console(parent, ns, title='Python'):
	existing = parent.get_data('PythonConsole')
	if existing:
		existing.present()
		return

	win = gtk.Window()

	parent._python_console = win

	win.set_transient_for(parent)
	win.set_destroy_with_parent(True)
	win.set_default_size(475, 300)
	win.set_title(title)

	def key_cb(win, event):
		if event.keyval in (gtk.keysyms.W, gtk.keysyms.w) and \
				event.state & gtk.gdk.CONTROL_MASK:
			win.hide()
		if event.keyval == gtk.keysyms.Escape:
			win.hide()
	win.connect('key_press_event', key_cb)

	def delete(win, event): win.hide(); return True
	win.connect('delete-event', delete)

	def quit(): win.hide()
	cons = Console(namespace=ns, quit_cb=quit)

	win.add(cons)
	cons.show()
	win.show()
	cons.init()
	return win

def console_cb(action, window):
	gtk_console(window, {
			'__builtins__':__builtins__,
			'window':window,
			})

import epiphany

_ui_str = """
<ui>
  <menubar name="menubar">
    <menu name="ToolsMenu" action="Tools">
      <separator/>
      <menuitem name="PythonConsole" action="PythonConsole"/>
      <separator/>
    </menu>
  </menubar>
</ui>
"""
_actions = [('PythonConsole', None, 'P_ython Console', None, None, console_cb)]

def attach_window(window):
	ui_manager = window.get_ui_manager()
	group = gtk.ActionGroup('PythonConsole')
	group.add_actions(_actions, window)
	ui_manager.insert_action_group(group, 0)
	ui_id = ui_manager.add_ui_from_string(_ui_str)

	window._python_console_window_data = (group, ui_id)

def detach_window(window):
	if hasattr(window, '_python_console'):
		window._python_console.destroy()
		del window._python_console

	group, ui_id = window._python_console_window_data
	del window._python_console_window_data

	ui_manager = window.get_ui_manager()
	ui_manager.remove_ui(ui_id)
	ui_manager.remove_action_group(group)
	ui_manager.ensure_update()



# -*- coding: utf-8 -*-

# Copyright (C) 2005 by Magnus Therning

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from sets import Set
import pickle
import os, os.path
import time

# Nice to have for debugging
logger = None
def get_logger():
    '''Create and return a file logger.

    The file used for logging is ~/.epilicious.log.

    @return: a file logger
    '''
    import logging
    global logger

    if logger:
        return logger

    logger = logging.getLogger('epilicious')
    hdlr = logging.FileHandler(os.path.join(os.environ['HOME'], '.epilicious.log'))
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(logging.ERROR)
    return logger

_file_name = os.path.join(os.environ['HOME'], '.gnome2', \
        'epiphany', 'epilicious_old')

def get_old():
    '''Read the base snapshot.

    The base snapshot is read from ~/.gnome2/epiphany/epilicious_old.

    @return: the base snapshot (an empty dictionary if none exists)
    '''
    res = {}
    try:
        f = open(_file_name, 'r')
        res = pickle.load(f)
        f.close()
    except IOError, e:
        res = {}
    return res

def save_snapshot(snap):
    '''Save a snapshot for future synchronisations.

    The base snapshot is written to ~/.gnome2/epiphany/epilicious_old.

    @param snap: The snapshot
    '''
    f = open(_file_name, 'w+')
    pickle.dump(snap, f)
    f.close()

def remove_urls(old, remote, local, rem_store, loc_store):
    '''Remove URLs from local and remote storage.

    @param old: The base snapshot
    @param remote: The remote snapshot
    @param local: The local snapshot
    @param rem_store: The remote storage (L{DeliciousStore})
    @param loc_store: The local storage (L{EpiphanyStore})
    '''
    o = Set(old.keys())
    r = Set(remote.keys())
    l = Set(local.keys())
    for url in o - l:
        rem_store.url_delete(url)
    for url in o - r:
        loc_store.url_delete(url)

def find_changed_urls(old, remote, local):
    '''Find the interesting set of URLs.

    A URL is interesting if it has been added or had its tags changed.

    >>> o = {0:'',     2:'',     4:'',     6:'',8:''}
    >>> r = {     1:'',2:'',     4:'',5:'',6:''     }
    >>> l = {0:'',     2:'',3:'',     5:'',6:''     }
    >>> res = calculate_pertinent_urls(o, r, l); res.sort(); res
    [0, 1, 3, 4]

    @param old: The base snapshot
    @param remote: The remote snapshot
    @param local: The local snapshot
    @return: a list of URLs
    '''

    def create_unique_id_for_url(url, desc, tags):
        tags.sort()
        return ','.join([url, desc, ','.join(tags)])

    o_ids = dict([[create_unique_id_for_url(u, old[u][0], old[u][1]), u] for u in old.keys()])
    r_ids = dict([[create_unique_id_for_url(u, remote[u][0], remote[u][1]), u] for u in remote.keys()])
    l_ids = dict([[create_unique_id_for_url(u, local[u][0], local[u][1]), u] for u in local.keys()])

    o = Set(o_ids.keys())
    r = Set(r_ids.keys())
    l = Set(l_ids.keys())

    l_changed = l - o
    r_changed = r - o

    result = []
    for id in l_changed:
        if l_ids[id] not in result: result.append(l_ids[id])
    for id in r_changed:
        if r_ids[id] not in result: result.append(r_ids[id])
    return result

def _get_tags(url, old, rem, loc):
    '''Get the tags and the description for a URL in all three locations.

    @param url: The URL to get tags for
    @param old: The base snapshot
    @param rem: The remote (Del.icio.us) snapshot
    @param loc: The local (Epiphany) snapshot
    @return: a tuple of description and lists of tags found in C{old}, C{rem},
        and C{loc}.
    '''
    try:
        otags = Set(old[url][1])
        desc = old[url][0]
    except:
        otags = Set()
    try:
        rtags = Set(rem[url][1])
        desc = rem[url][0]
    except:
        rtags = Set()
    try:
        ltags = Set(loc[url][1])
        desc = loc[url][0]
    except:
        ltags = Set()
    return desc, otags, rtags, ltags

def sync_tags_on_urls(curls, old, remote, \
        local, rem_store, loc_store):
    '''Synchronise tags on URLs, and add URLs that don't exist.

    @param curls: The list of URLs to synchronise
    @param old: The base snapshot
    @param remote: The remote snapshot
    @param local: The local snapshot
    @param rem_store: The remote storage (L{DeliciousStore})
    @param loc_store: The local storage (L{EpiphanyStore})
    '''
    for url in curls:
        desc, otags, rtags, ltags = _get_tags(url, old, remote, local)
        if otags and (not rtags or not ltags):
            continue
        rem_store.url_sync(url, desc, otags - ltags, ltags - otags)
        loc_store.url_sync(url, desc, otags - rtags, rtags - otags)



# -*- coding: utf-8 -*-

# Copyright (C) 2006 by Magnus Therning

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import libepilicious
import gtk, gtk.glade
import gconf
import os

class ConfigWindow:

    def __init__(self):
        self.gui = gtk.glade.XML(os.path.join(libepilicious.GLADE_DIR, \
                'epilicious-config.glade'), domain='epilicious')
        self.gui.signal_autoconnect(self)

        # I can't remember how to get the whole friggin' lot of 'em...
        self.config = self.gui.get_widget('dlgConfig')
        self.entname = self.gui.get_widget('entName')
        self.entpassword = self.gui.get_widget('entPassword')
        self.enttag = self.gui.get_widget('entTag')
        self.rbinclude = self.gui.get_widget('rbInclude')
        self.rbexclude = self.gui.get_widget('rbExclude')
        #self.btnhelp = self.gui.get_widget('btnHelp')
        self.btnclose = self.gui.get_widget('btnClose')
        self.client = gconf.client_get_default()

        # For some reason this gets messed up if set in Glade
        self.entpassword.set_visibility(False)
        self.entpassword.set_invisible_char('●')

    def show(self):
        self.entname.set_text(self.client.get_string('/apps/epiphany/extensions/epilicious/username'))
        self.entpassword.set_text(self.client.get_string('/apps/epiphany/extensions/epilicious/password'))
        self.enttag.set_text(self.client.get_string('/apps/epiphany/extensions/epilicious/keyword'))
        self.rbexclude.set_active(self.client.get_bool('/apps/epiphany/extensions/epilicious/exclude'))

    def hide(self):
        self.config.hide()
        self.client.set_string('/apps/epiphany/extensions/epilicious/username', self.entname.get_text())
        self.client.set_string('/apps/epiphany/extensions/epilicious/password', self.entpassword.get_text())
        self.client.set_string('/apps/epiphany/extensions/epilicious/keyword', self.enttag.get_text())
        self.client.set_bool('/apps/epiphany/extensions/epilicious/exclude', self.rbexclude.get_active())
        return False

    ### signal handlers
    def on_entName_focus_out_event(self, widget, event):
        self.client.set_string('/apps/epiphany/extensions/epilicious/username', self.entname.get_text())

    def on_entPassword_focus_out_event(self, widget, event):
        self.client.set_string('/apps/epiphany/extensions/epilicious/password', self.entpassword.get_text())

    def on_entTag_focus_out_event(self, widget, event):
        self.client.set_string('/apps/epiphany/extensions/epilicious/keyword', self.enttag.get_text())

    def on_rb_clicked(self, widget):
        self.client.set_bool('/apps/epiphany/extensions/epilicious/exclude', self.rbexclude.get_active())

    def on_btnHelp_clicked(self, widget):
        # Figure out how to open help and open it to the yet-unwritten
        # Epilicious help file. :)
        pass

    def on_btnClose_clicked(self, widget):
        self.hide()
        del(self)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 by Magnus Therning

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from sets import Set
import epiphany
import libepilicious
from libepilicious.BaseStore import BaseStore


class EpiphanyStore(BaseStore):

    def __init__(self, keyword = None, exclude = False):
        # TODO: add a check for the priority of the keyword and set it to 0, at
        # the moment it's not possible since there's no method for it
        # (ephy_node_set_property() isn't in the Python API)
        self.__kw = keyword
        self.__excl = exclude

        self.__bms = epiphany.ephy_shell_get_default().get_bookmarks()
        if not self.__bms.find_keyword(self.__kw, False):
            self.__bms.add_keyword(self.__kw)

    def __get_all_shared_bookmarks(self):
        # Return all shared bookmarks
        # Two possibilities:
        #  1. Take all bookmarks we have and remove the ones in __kw
        #  2. All bookmarks are in __kw
        key = self.__bms.find_keyword(self.__kw, False)
        if self.__excl:
            tmp = self.__bms.get_bookmarks().get_children()
            res = []
            for b in tmp:
                if not self.__bms.has_keyword(key, b):
                    res.append(b)
            return res
        else:
            return key.get_children()

    def __get_all_keywords(self):
        # Return a list containing all interesting keywords. All special
        # keywords (priority==1) and 'All' (id==0) are uninteresting.
        kw_node = self.__bms.find_keyword(self.__kw, False)
        return [kw for kw in self.__bms.get_keywords().get_children() \
                if kw.get_id() != 0 and \
                kw.get_property_int(epiphany.NODE_KEYWORD_PROP_PRIORITY) != 1 and \
                kw != kw_node]

    def get_snapshot(self):
        '''Calculates a snapshot of the Epiphany bookmarks.

        @note: L{BaseStore.get_snapshot} documents the format of the return
        value.
        '''
        all_shared_bms = self.__get_all_shared_bookmarks()
        all_keywords = self.__get_all_keywords()

        res = {}
        for bm in all_shared_bms:
            res[bm.get_property_string(epiphany.NODE_BMK_PROP_LOCATION)] = \
                    [bm.get_property_string(epiphany.NODE_BMK_PROP_TITLE)]
            keywords = []
            for key in all_keywords:
                if self.__bms.has_keyword(key, bm):
                    keywords.append(key.get_property_string(epiphany.NODE_KEYWORD_PROP_NAME))
            res[bm.get_property_string(epiphany.NODE_BMK_PROP_LOCATION)].append(keywords)

        return res

    def url_delete(self, url):
        '''Delete a bookmark.

        :type url: string
        :param url: The URL of the bookmark
        '''
        # This is brute force. It seems the easiest way of removing a bookmark
        # is by removing every keyword there is from it.
        bm = self.__bms.find_bookmark(url)
        if bm:
            for kw in self.__bms.get_keywords().get_children():
                self.__bms.unset_keyword(kw, bm)

    def url_sync(self, url, desc, to_del, to_add):
        '''Synchronise a bookmark.

        The bookmark with the given URL is created if needed and its keywords
        are adjusted.

        :type url: string
        :param url: The URL of the bookmark
        :type desc: string
        :param desc: One-line description of the bookmark
        :type to_del: list
        :param to_del: The keywords (as strings) to delete from the bookmark
        :type to_add: list
        :param to_add: The keywords (as strings) to add to the bookmark
        '''
        bm = self.__bms.find_bookmark(url)
        sharekw = self.__bms.find_keyword(self.__kw, False)
        if not bm:
            bm = self.__bms.add(desc, url)

        if self.__excl:
            if self.__bms.has_keyword(sharekw, bm):
                self.__bms.unset_keyword(sharekw, bm)
        else:
            if not self.__bms.has_keyword(sharekw, bm):
                self.__bms.set_keyword(sharekw, bm)

        for k in to_del:
            kw = self.__bms.find_keyword(k, False)
            self.__bms.unset_keyword(kw, bm)

        for k in to_add:
            kw = self.__bms.find_keyword(k, False)
            if not kw:
                kw = self.__bms.add_keyword(k)
            self.__bms.set_keyword(kw, bm)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 by Magnus Therning

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import gtk
import gconf
import gobject
import sys, os, os.path

if sys.version_info[0:2] < (2, 5):
    class GeneratorExit(Exception):
        pass

sys.path.append('/usr/lib/epiphany/2.17/extensions')

import libepilicious
libepilicious.GLADE_DIR = '/usr/local/share/epiphany-extensions/glade'

# localization
import gettext
try:
    t = gettext.translation('epilicious')
    _ = t.ugettext
except Exception, e:
    _ = lambda x : x

## Globals ##
username = ''
password = ''
keyword = ''
exclude = False
space_repl = ''

## Configuration ##
_gconf_dir = '/apps/epiphany/extensions/epilicious'
_gconf_un = '/apps/epiphany/extensions/epilicious/username'
_gconf_pwd = '/apps/epiphany/extensions/epilicious/password'
_gconf_kw = '/apps/epiphany/extensions/epilicious/keyword'
_gconf_excl = '/apps/epiphany/extensions/epilicious/exclude'

def _new_un(client, *args, **kwargs):
    '''Callback to handle the username is modified in GConf.'''
    global username
    username = client.get_string(_gconf_un)

def _new_pwd(client, *args, **kwargs):
    '''Callback to handle the password is modified in GConf.'''
    global password
    password = client.get_string(_gconf_pwd)

def _new_keyword(client, *args, **kwargs):
    '''Callback to handle the keyword is modified in GConf.'''
    global keyword
    keyword = client.get_string(_gconf_kw)

def _new_exclude(client, *args, **kwargs):
    '''Callback to handle the exclude is modified in GConf.'''
    global exclude
    exclude = client.get_bool(_gconf_excl)

def _gconf_register():
    '''Sets up the GConf callbacks.'''
    global keyword, exclude, space_repl

    # hard coded for now
    space_repl = '#'

    client = gconf.client_get_default()
    client.add_dir(_gconf_dir, gconf.CLIENT_PRELOAD_NONE)
    client.notify_add(_gconf_un, _new_un)
    client.notify_add(_gconf_pwd, _new_pwd)
    client.notify_add(_gconf_kw, _new_keyword)
    client.notify_add(_gconf_excl, _new_exclude)
    _new_un(client)
    _new_pwd(client)
    _new_keyword(client)
    _new_exclude(client)

_gconf_register()

## Synchronisation ##
def _CB_Sync(action, window):
    # Using gobject.idle_add() together with a generator is much easier than
    # getting threading to work properly. It doesn't make the GUI very
    # responsive during del.icio.us calls, but it's better than nothing.
    sync_gen = _do_sync()
    gobject.idle_add(sync_gen.next)

def _CB_Config(action, window):
    from libepilicious.config import ConfigWindow
    config = ConfigWindow()
    config.show()

def _do_sync():
    '''Perform the synchronisation.

    This is the method called from the menu item added in epiphany. The
    algorithm is as follows:

    1. Read the base snapshot (L{libepilicious.get_old})
    2. Create remote storage representation (L{libepilicious.DeliciousStore})
    3. Create local storage representation (L{libepilicious.EpiphanyStore})
    4. Remove URLs (L{libepilicous.remove_urls})
    5. Add new URLs and synchronise tags (L{libepilicous.sync_tags_on_urls})
    6. Save a local snapshot as a base for the next synchronisation
       (L{libepilicious.save_snapshot})

    Any errors that occur are logged.
    '''
    from libepilicious.progress import ProgressBar
    pbar = ProgressBar()
    try:
        from libepilicious import *
        from libepilicious.DeliciousStore import DeliciousStore
        from libepilicious.EpiphanyStore import EpiphanyStore

        pbar.show()
        stepper = pbar.step()

        remote_store = DeliciousStore(user = username, pwd = password, \
                space_repl = space_repl)
        local_store = EpiphanyStore(keyword = keyword, exclude = exclude)
        stepper.next()
        old = get_old()
        #libepilicious.get_logger().info('old=' + str(old))
        yield True
        stepper.next()
        remote = remote_store.get_snapshot()
        #libepilicious.get_logger().info('remote=' + str(remote))
        yield True
        stepper.next()
        local = local_store.get_snapshot()
        #libepilicious.get_logger().info('local=' + str(local))
        yield True
        stepper.next()

        # Synchronize URLs
        remove_urls(old = old,
                remote = remote,
                local = local,
                rem_store = remote_store,
                loc_store = local_store)
        yield True
        stepper.next()

        changed_urls = find_changed_urls(old = old,
                remote = remote, local = local)
        #libepilicious.get_logger().info('changed_urls=' + str(changed_urls))
        sync_tags_on_urls(curls = changed_urls,
                old = old,
                remote = remote,
                local = local,
                rem_store = remote_store,
                loc_store = local_store)
        yield True
        stepper.next()

        # Save the current state for future sync
        save_snapshot(local_store.get_snapshot())
        yield True
        stepper.next()
        yield False
    except StopIteration, e:
        libepilicious.get_logger().exception('Too many calls to stepper.next()')
    except GeneratorExit, e:
        raise e
    except:
        pbar.failed()
        libepilicious.get_logger().exception('Failed sync')


## Epiphany integration ##

_menu_ui = '''
<ui>
  <menubar name="menubar">
    <menu name="BookmarksMenu" action="Bookmarks">
      <separator />
      <menuitem name="%s" action="EpiliciousSync" />
      <separator />
    </menu>
    <menu name="ToolsMenu" action="Tools">
      <menuitem name="EpiliciousConfig" action="EpiliciousConfig"/>
    </menu>
  </menubar>
</ui>
''' % (_('Epilicious Synchronize'))

_actions = [ \
        ('EpiliciousSync', None, _('Epilicious Synchronize'),
                None, None, _CB_Sync), \
        ('EpiliciousConfig', None, ('Epilicious Configuration'),
                None, None, _CB_Config), \
        ]

# Epiphany extension interface
def attach_window(window):
    try:
        ui_manager = window.get_ui_manager()
        group = gtk.ActionGroup('My Menu')
        group.add_actions(_actions, window)
        ui_manager.insert_action_group(group, 0)
        ui_id = ui_manager.add_ui_from_string(_menu_ui)

        window._my_menu_data = (group, ui_id)
    except Exception, e:
        libepilicious.get_logger().exception('Failed attach')

def detach_window(window):
    try:
        group, ui_id = window._my_menu_data
        del window._my_menu_data

        ui_manager = window.get_ui_manager()
        ui_manager.remove_ui(ui_id)
        ui_manager.remove_action_group(group)
        ui_manager.ensure_update()
    except Exception, e:
        libepilicious.get_logger().exception('Failed detach')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Sample Python Menu extension: creates menu entries using Python
#
# Copyright © 2005 Adam Hooper
#
# Places two menu entries in the Tools menu: one which is always enabled, and
# one which will only be enabled when the currently-viewed page has finished
# loading.

import gtk

_ui_str = """
<ui>
 <menubar name="menubar">
  <menu name="ToolsMenu" action="Tools">
   <separator/>
   <menuitem name="PythonSampleMenuAlwaysEnabled"
             action="PythonSampleMenuAlwaysEnabled"/>
   <menuitem name="PythonSampleMenuSometimesEnabled"
	     action="PythonSampleMenuSometimesEnabled"/>
   <separator/>
  </menu>
 </menubar>
</ui>
"""

# These are the callbacks: what our menu entries will do
def _dialog(window, msg):
	dialog = gtk.MessageDialog(window, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO,
				   gtk.BUTTONS_OK, msg)
	dialog.run()
	dialog.destroy()

def _always_cb(action, window):
	_dialog(window, 'You clicked a menu entry')

def _sometimes_cb(action, window):
	_dialog(window, 'If you see this, the page must be loaded')

# This is to pass to gtk.ActionGroup.add_actions()
_actions = [('PythonSampleMenuAlwaysEnabled', None,
	     'Python Sample (always enabled)', None, None, _always_cb),
	    ('PythonSampleMenuSometimesEnabled', None,
	     'Python Sample (sometimes enabled)', None, None, _sometimes_cb),
	   ]

# The next three functions handle the SometimesEnabled action. Note that we
# use window.get_active_tab(): we want the menu entries to reflect the active
# tab, not necessarily the one which fired a signal.
def _update_action(window):
	sometimes = window.get_ui_manager().get_action('/menubar/ToolsMenu/PythonSampleMenuSometimesEnabled')

	tab = window.get_active_tab()

	# Tab is None when a window is first opened
	sensitive = (tab != None and tab.get_load_status() != True)
	sometimes.set_sensitive(sensitive)

def _switch_page_cb(notebook, page, page_num, window):
	_update_action(window)

def _load_status_cb(tab, pspec, window):
	_update_action(window)

# These implement the EphyExtension interface.
#
# The tab_attached, tab_detached and notebook parts are only necessary for the
# SometimesEnabled action. In other words, no signals are needed for the
# AlwaysEnabled menu entry.
def attach_window(window):
	ui_manager = window.get_ui_manager()
	group = gtk.ActionGroup('PythonSampleMenu')
	group.add_actions(_actions, window)
	ui_manager.insert_action_group(group, -1)
	ui_id = ui_manager.add_ui_from_string(_ui_str)

	window._python_sample_menu_window_data = (group, ui_id)

	notebook = window.get_notebook()
	sig = notebook.connect('switch_page', _switch_page_cb, window)
	notebook._python_sample_menu_sig = sig

def detach_window(window):
	notebook = window.get_notebook()
	notebook.disconnect(notebook._python_sample_menu_sig)
	del notebook._python_sample_menu_sig

	group, ui_id = window._python_sample_menu_window_data
	del window._python_sample_menu_window_data

	ui_manager = window.get_ui_manager()
	ui_manager.remove_ui(ui_id)
	ui_manager.remove_action_group(group)
	ui_manager.ensure_update()

def attach_tab(window, tab):
	sig = tab.connect('notify::load-status', _load_status_cb, window)
	tab._python_sample_menu_sig = sig

def detach_tab(window, tab):
	tab.disconnect(tab._python_sample_menu_sig)
	del tab._python_sample_menu_sig



# -*- coding: utf-8 -*-

# Copyright (C) 2005 by Magnus Therning

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

class BaseStore:
    '''The base of all stores. Never to be instantiated.'''

    def get_snapshot(self):
        '''Calculates a snapshot of the store's bookmarks.

        The format of the snapshot is::
         { <url> : [ <description>, [tag, tag, ...]],
           <url> : [ <description>, [<tags>*]],
           ... }

        @return: the snapshot
        '''
        pass


    def url_delete(self, url):
        '''Deletes a URL form the store.

        @type url: string
        @param url: URL to delete
        '''
        pass


    def url_sync(self, url, desc, to_del, to_add):
        '''Synchronises a URL's tags. The URL is added if it doesn't exist.

        @param url: The URL
        @param to_del: Set of tags to delete
        @param to_add: Set of tags to add
        '''
        pass



# -*- coding: utf-8 -*-

# Copyright (C) 2005 by Magnus Therning

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from urllib2 import URLError
from sets import Set
import delicious
import libepilicious
from libepilicious.BaseStore import BaseStore


class DeliciousStore(BaseStore):
    '''The class representing the storage of Delicious bookmarks.'''

    def __init__(self, user, pwd, space_repl):
        '''Constructor.

        @param user: Delicious username
        @param pwd: Delicious password
        @param space_repl: Character that replaces space
        '''
        self.__un = user
        self.__pwd = pwd
        self.__sr = space_repl
        self.__d = delicious.Delicious(user, pwd)
        self.__snap_utd = 0

    def get_snapshot(self):
        '''Calculates a snapshot of the del.icio.us bookmarks.

        @note: L{BaseStore.get_snapshot} documents the format of the return
        value.
        '''
        if self.__snap_utd:
            return self.__snap
        all = self.__d.all_bookmarks()
        res = {}
        for p in all:
            # Delicious prepends "dangerous" links
            if p['href'][:33] == 'http://del.icio.us/doc/dangerous#':
                url = p['href'][33:]
            else:
                url = p['href']
            res[url] = [p['description'], \
                    [t.replace(self.__sr, ' ') for t in p['tag'].split(' ')]]

        self.__snap = res
        self.__snap_utd = 1
        return res

    def url_delete(self, url):
        '''Deletes a URL from the storage.

        @param url: The URL to delete
        '''
        # Delicious prepends "dangerous" links
        if url[:7] == 'file://':
            url = 'http://del.icio.us/doc/dangerous#' + url
        try:
            self.__d.delete_bookmark(url)
            self.__snap_utd = 0
        except:
            libepilicious.get_logger().exception('Failed to delete URL %s' % url)

    def url_sync(self, url, desc, to_del, to_add):
        '''Synchronises a URL's tags. The URL is added if it doesn't exist.

        @param url: The URL
        @param to_del: Set of tags to delete
        @param to_add: Set of tags to add
        '''
        if to_del or to_add:
            snap = self.get_snapshot()
            if snap.has_key(url):
                tags = (Set(snap[url][1]) | to_add) - to_del
            else:
                tags = to_add
            tag_str = ' '.join([t.replace(' ', self.__sr) for t in tags])
            try:
                self.__d.add_bookmark(url, description=desc, \
                        tags=tag_str)
                self.__snap_utd = 0
            except:
                libepilicious.get_logger().exception('Failed to synchronise URL %s' % url)
